import ctypes
from enum import Enum

from . import sdk
from .enum import (EntitlementType, ImageType, InputModeType, LobbyType,
                   RelationshipType, SkuType, Status)


class Model:
    def __init__(self, **kwargs):
        self._internal = kwargs.get("internal", self._struct_())
        if "copy" in kwargs:
            ctypes.memmove(
                ctypes.byref(self._internal),
                ctypes.byref(kwargs["copy"]),
                ctypes.sizeof(self._struct_)
            )

        self._fields = {}

        for name, field, ftype in self._fields_:
            self._fields[name] = (field, ftype)
            if issubclass(ftype, Model):
                setattr(self, "_" + field, ftype(internal=getattr(self._internal, field)))

    def __getattribute__(self, key):
        if key.startswith("_"):
            return super().__getattribute__(key)
        else:
            field = self._fields[key]
            value = getattr(self._internal, field[0])
            if field[1] == int:
                return int(value)
            elif field[1] == str:
                return value.decode("utf8")
            elif field[1] == bool:
                return bool(value)
            elif issubclass(field[1], Model):
                return getattr(self, "_" + field[0])
            elif issubclass(field[1], Enum):
                return field[1](int(value))
            else:
                raise TypeError(field[1])

    def __setattr__(self, key, value):
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            field = self._fields[key]
            if field[1] == int:
                value = int(value)
                setattr(self._internal, field[0], value)
            elif field[1] == str:
                value = value.encode("utf8")
                setattr(self._internal, field[0], value)
            elif field[1] == bool:
                value = bool(value)
                setattr(self._internal, field[0], value)
            elif issubclass(field[1], Model):
                setattr(self, "_" + field[0], value)
                setattr(self._internal, field[0], value._internal)
            elif issubclass(field[1], Enum):
                setattr(self._internal, field[0], value.value)
            else:
                raise TypeError(field[1])

    def __dir__(self):
        return super().__dir__() + list(self._fields.keys())


class User(Model):
    _struct_ = sdk.DiscordUser
    _fields_ = [
        ("Id", "id", int),
        ("Username", "username", str),
        ("Discriminator", "discriminator", str),
        ("Avatar", "avatar", str),
        ("Bot", "bot", bool),
    ]


class ActivityTimestamps(Model):
    _struct_ = sdk.DiscordActivityTimestamps
    _fields_ = [
        ("Start", "start", int),
        ("End", "end", int),
    ]


class ActivityAssets(Model):
    _struct_ = sdk.DiscordActivityAssets
    _fields_ = [
        ("LargeImage", "large_image", str),
        ("LargeText", "large_text", str),
        ("SmallImage", "small_image", str),
        ("SmallText", "small_text", str),
    ]


class PartySize(Model):
    _struct_ = sdk.DiscordPartySize
    _fields_ = [
        ("CurrentSize", "current_size", int),
        ("MaxSize", "max_size", int),
    ]


class ActivityParty(Model):
    _struct_ = sdk.DiscordActivityParty
    _fields_ = [
        ("Id", "id", str),
        ("Size", "size", PartySize),
    ]


class ActivitySecrets(Model):
    _struct_ = sdk.DiscordActivitySecrets
    _fields_ = [
        ("Match", "match", str),
        ("Join", "join", str),
        ("Spectate", "spectate", str),
    ]


class Activity(Model):
    _struct_ = sdk.DiscordActivity
    _fields_ = [
        ("ApplicationId", "application_id", int),
        ("Name", "name", str),
        ("State", "state", str),
        ("Details", "details", str),
        ("Timestamps", "timestamps", ActivityTimestamps),
        ("Assets", "assets", ActivityAssets),
        ("Party", "party", ActivityParty),
        ("Secrets", "secrets", ActivitySecrets),
        ("Instance", "instance", bool),
    ]


class Presence(Model):
    _struct_ = sdk.DiscordPresence
    _fields_ = [
        ("Status", "status", Status),
        ("Activity", "activity", Activity),
    ]


class Relationship(Model):
    _struct_ = sdk.DiscordRelationship
    _fields_ = [
        ("Type", "type", RelationshipType),
        ("User", "user", User),
        ("Presence", "presence", Presence),
    ]


class ImageDimensions(Model):
    _struct_ = sdk.DiscordImageDimensions
    _fields_ = [
        ("Width", "width", int),
        ("Height", "height", int),
    ]


class ImageHandle(Model):
    _struct_ = sdk.DiscordImageHandle
    _fields_ = [
        ("Type", "type", ImageType),
        ("Id", "id", int),
        ("Size", "size", int),
    ]


class OAuth2Token(Model):
    _struct_ = sdk.DiscordOAuth2Token
    _fields_ = [
        ("AccessToken", "access_token", str),
        ("Scopes", "scopes", str),
        ("Expires", "expires", int),
    ]


class Lobby(Model):
    _struct_ = sdk.DiscordLobby
    _fields_ = [
        ("Id", "id", int),
        ("Type", "type", LobbyType),
        ("OwnerId", "owner_id", int),
        ("Secret", "secret", str),
        ("Capacity", "capacity", int),
        ("Locked", "locked", bool),
    ]


class InputMode(Model):
    _struct_ = sdk.DiscordInputMode
    _fields_ = [
        ("Type", "type", InputModeType),
        ("Shortcut", "shortcut", str),
    ]


class FileStat(Model):
    _struct_ = sdk.DiscordFileStat
    _fields_ = [
        ("Filename", "filename", str),
        ("Size", "size", int),
        ("LastModified", "last_modified", int),
    ]


class UserAchievement(Model):
    _struct_ = sdk.DiscordUserAchievement
    _fields_ = [
        ("UserId", "user_id", str),
        ("AchievementId", "achievement_id", int),
        ("PercentComplete", "percent_complete", int),
        ("UnlockedAt", "unlocked_at", str),
    ]


class SkuPrice(Model):
    _struct_ = sdk.DiscordSkuPrice
    _fields_ = [
        ("Amount", "amount", int),
        ("Currency", "currency", str),
    ]


class Sku(Model):
    _struct_ = sdk.DiscordSku
    _fields_ = [
        ("Id", "id", int),
        ("Type", "type", SkuType),
        ("Name", "name", str),
        ("Price", "price", SkuPrice),
    ]


class Entitlement(Model):
    _struct_ = sdk.DiscordEntitlement
    _fields_ = [
        ("Id", "id", int),
        ("Type", "type", EntitlementType),
        ("SkuId", "sku_id", int),
    ]



import ctypes
from typing import Callable

from . import sdk
from .enum import ActivityActionType, Result
from .event import bindEvents


class OverlayManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordOverlayEvents,
            self._OnToggle
        )

    def _OnToggle(self, event_data, locked):
        self.OnToggle(locked)

    def IsEnabled(self) -> bool:
        """
        Check whether the user has the overlay enabled or disabled.
        """
        enabled = ctypes.c_bool()
        self._internal.is_enabled(self._internal, enabled)
        return enabled.value

    def IsLocked(self) -> bool:
        """
        Check if the overlay is currently locked or unlocked
        """
        locked = ctypes.c_bool()
        self._internal.is_locked(self._internal, locked)
        return locked.value

    def SetLocked(self, locked: bool, callback: Callable[[Result], None]) -> None:
        """
        Locks or unlocks input in the overlay.
        """
        def CCallback(event_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.set_locked.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.set_locked(self._internal, locked, ctypes.c_void_p(), CCallback)

    def OpenActivityInvite(
        self,
        type: ActivityActionType,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Opens the overlay modal for sending game invitations to users, channels, and servers.
        """
        def CCallback(event_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.open_activity_invite.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.open_activity_invite(self._internal, type, ctypes.c_void_p(), CCallback)

    def OpenGuildInvite(self, code: str, callback: Callable[[Result], None]) -> None:
        """
        Opens the overlay modal for joining a Discord guild, given its invite code.
        """
        def CCallback(event_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.open_guild_invite.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        code = ctypes.c_char_p(code.encode("utf8"))
        self._internal.open_guild_invite(self._internal, code, ctypes.c_void_p(), CCallback)

    def OpenVoiceSettings(self, callback: Callable[[Result], None]) -> None:
        """
        Opens the overlay widget for voice settings for the currently connected application.
        """
        def CCallback(event_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.open_voice_settings.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.open_voice_settings(self._internal, ctypes.c_void_p(), CCallback)

    def OnToggle(self, locked: bool) -> None:
        """
        Fires when the overlay is locked or unlocked (a.k.a. opened or closed)
        """
        pass



import ctypes
from typing import Callable

from . import sdk
from .enum import ActivityActionType, ActivityJoinRequestReply, Result
from .event import bindEvents
from .model import Activity, User


class ActivityManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordActivityEvents,
            self._OnActivityJoin,
            self._OnActivitySpectate,
            self._OnActivityJoinRequest,
            self._OnActivityInvite
        )

    def _OnActivityJoin(self, event_data, secret):
        self.OnActivityJoin(secret.decode("utf8"))

    def _OnActivitySpectate(self, event_data, secret):
        self.OnActivitySpectate(secret.decode("utf8"))

    def _OnActivityJoinRequest(self, event_data, user):
        self.OnActivityJoinRequest(User(copy=user.contents))

    def _OnActivityInvite(self, event_data, type, user, activity):
        self.OnActivityInvite(type, User(copy=user.contents), Activity(copy=activity.contents))

    def RegisterCommand(self, command: str) -> Result:
        """
        Registers a command by which Discord can launch your game.
        """
        result = Result(self._internal.register_command(self._internal, command.encode("utf8")))
        return result

    def RegisterSteam(self, steamId: int) -> Result:
        """
        Registers your game's Steam app id for the protocol `steam://run-game-id/<id>`.
        """
        result = Result(self._internal.register_steam(self._internal, steamId))
        return result

    def UpdateActivity(self, activity: Activity, callback: Callable[[Result], None]) -> None:
        """
        Set a user's presence in Discord to a new activity.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.update_activity.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.update_activity(
            self._internal, activity._internal, ctypes.c_void_p(), CCallback)

    def ClearActivity(self, callback: Callable[[Result], None]) -> None:
        """
        Clears a user's presence in Discord to make it show nothing.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.clear_activity.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.clear_activity(self._internal, ctypes.c_void_p(), CCallback)

    def SendRequestReply(
        self,
        userId: int,
        reply: ActivityJoinRequestReply,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Sends a reply to an Ask to Join request.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.send_request_reply.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.send_request_reply(
            self._internal,
            userId,
            reply,
            ctypes.c_void_p(),
            CCallback
        )

    def SendInvite(
        self,
        userId: int,
        type: ActivityActionType,
        content: str,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Sends a game invite to a given user.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.send_invite.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.send_invite(
            self._internal,
            userId,
            type,
            content.encode("utf8"),
            ctypes.c_void_p(),
            CCallback
        )

    def AcceptInvite(self, userId: int, callback: Callable[[Result], None]) -> None:
        """
        Accepts a game invitation from a given userId.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.accept_invite.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.accept_invite(self._internal, userId, ctypes.c_void_p(), CCallback)

    def OnActivityJoin(self, joinSecret: str) -> None:
        """
        Fires when a user accepts a game chat invite or receives confirmation from Asking to Join.
        """
        pass

    def OnActivitySpectate(self, spectateSecret: str) -> None:
        """
        Fires when a user accepts a spectate chat invite or clicks the Spectate button on a user's
        profile.
        """
        pass

    def OnActivityJoinRequest(self, user: User) -> None:
        """
        Fires when a user asks to join the current user's game.
        """
        pass

    def OnActivityInvite(self, type: ActivityActionType, user: User, activity: Activity) -> None:
        """
        Fires when the user receives a join or spectate invite.
        """
        pass



import ctypes


def bindEvents(structure, *methods):
    contents = structure()
    for index, (name, func) in enumerate(structure._fields_):
        setattr(contents, name, func(methods[index]))

    pointer = ctypes.pointer(contents)
    return pointer



from .enum import Result


class DiscordException(Exception):
    pass


exceptions = {}

# we dynamically create the exceptions
for res in Result:
    exception = type(res.name, (DiscordException,), {})

    globals()[res.name] = exception
    exceptions[res] = exception


def getException(result):
    return exceptions.get(result, DiscordException)("result " + str(result.value))



import ctypes
from typing import Callable

from . import sdk
from .enum import Result
from .event import bindEvents
from .exception import getException
from .model import InputMode


class VoiceManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordVoiceEvents,
            self._OnSettingsUpdate
        )

    def _OnSettingsUpdate(self, event_data):
        self.OnSettingsUpdate()

    def GetInputMode(self) -> InputMode:
        """
        Get the current voice input mode for the user
        """
        input_mode = sdk.DiscordInputMode()
        result = Result(self._internal.get_input_mode(self._internal, input_mode))
        if result != Result.Ok:
            raise getException(result)

        return InputMode(internal=input_mode)

    def SetInputMode(self, inputMode: InputMode, callback: Callable[[Result], None]) -> None:
        """
        Sets a new voice input mode for the uesr.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.set_input_mode.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.set_input_mode(
            self._internal,
            inputMode._internal,
            ctypes.c_void_p(),
            CCallback
        )

    def IsSelfMute(self) -> bool:
        """
        Whether the connected user is currently muted.
        """
        mute = ctypes.c_bool()
        result = Result(self._internal.is_self_mute(self._internal, mute))
        if result != Result.Ok:
            raise getException(result)

        return mute.value

    def SetSelfMute(self, mute: bool) -> None:
        """
        Mutes or unmutes the currently connected user.
        """
        result = Result(self._internal.set_self_mute(self._internal, mute))
        if result != Result.Ok:
            raise getException(result)

    def IsSelfDeaf(self) -> bool:
        """
        Whether the connected user is currently deafened.
        """
        deaf = ctypes.c_bool()
        result = Result(self._internal.is_self_deaf(self._internal, deaf))
        if result != Result.Ok:
            raise getException(result)

        return deaf.value

    def SetSelfDeaf(self, deaf: bool) -> None:
        """
        Deafens or undefeans the currently connected user.
        """
        result = Result(self._internal.set_self_deaf(self._internal, deaf))
        if result != Result.Ok:
            raise getException(result)

    def IsLocalMute(self, userId: int) -> bool:
        """
        Whether the given user is currently muted by the connected user.
        """
        mute = ctypes.c_bool()
        result = Result(self._internal.is_local_mute(self._internal, userId, mute))
        if result != Result.Ok:
            raise getException(result)

        return mute.value

    def SetLocalMute(self, userId: int, mute: bool) -> None:
        """
        Mutes or unmutes the given user for the currently connected user.
        """
        result = Result(self._internal.set_local_mute(self._internal, userId, mute))
        if result != Result.Ok:
            raise getException(result)

    def GetLocalVolume(self, userId: int) -> int:
        """
        Gets the local volume for a given user.
        """
        volume = ctypes.c_uint8()
        result = Result(self._internal.get_local_volume(self._internal, userId, volume))
        if result != Result.Ok:
            raise getException(result)

        return volume.value

    def SetLocalVolume(self, userId: int, volume: int) -> None:
        """
        Sets the local volume for a given user.
        """
        result = Result(self._internal.set_local_volume(self._internal, userId, volume))
        if result != Result.Ok:
            raise getException(result)

    def OnSettingsUpdate(self) -> None:
        # This event is not documented anywhere (yet?)
        pass



from . import sdk
from .achievement import AchievementManager
from .activity import ActivityManager
from .application import ApplicationManager, SignedAppTicket
from .discord import Discord
from .enum import (
    ActivityActionType, ActivityJoinRequestReply, ActivityType, CreateFlags,
    EntitlementType, ImageType, InputModeType, LobbySearchCast,
    LobbySearchComparison, LobbySearchDistance, LobbyType, LogLevel,
    PremiumType, RelationshipType, Result, SkuType, Status, UserFlag)
from .event import bindEvents
from .exception import DiscordException, exceptions, getException
from .image import ImageManager
from .lobby import (LobbyManager, LobbyMemberTransaction, LobbySearchQuery,
                    LobbyTransaction)
from .model import (
    Activity, ActivityAssets, ActivityParty, ActivitySecrets,
    ActivityTimestamps, Entitlement, FileStat, ImageDimensions, ImageHandle,
    InputMode, Lobby, Model, OAuth2Token, PartySize, Presence, Relationship,
    Sku, SkuPrice, User, UserAchievement)
from .network import NetworkManager
from .overlay import OverlayManager
from .relationship import RelationshipManager
from .storage import StorageManager
from .store import StoreManager
from .user import UserManager
from .voice import VoiceManager

__all__ = [
    "sdk",
    "AchievementManager",
    "ActivityManager",
    "ApplicationManager", "SignedAppTicket",
    "Discord",
    "ActivityActionType", "ActivityJoinRequestReply", "ActivityType", "CreateFlags",
    "EntitlementType", "ImageType", "InputModeType", "LobbySearchCast",
    "LobbySearchComparison", "LobbySearchDistance", "LobbyType", "LogLevel",
    "PremiumType", "RelationshipType", "Result", "SkuType", "Status", "UserFlag",
    "bindEvents",
    "DiscordException", "exceptions", "getException",
    "ImageManager",
    "LobbyManager", "LobbyMemberTransaction", "LobbySearchQuery",
    "LobbyTransaction",
    "Activity", "ActivityAssets", "ActivityParty", "ActivitySecrets",
    "ActivityTimestamps", "Entitlement", "FileStat", "ImageDimensions", "ImageHandle",
    "InputMode", "Lobby", "Model", "OAuth2Token", "PartySize", "Presence", "Relationship",
    "Sku", "SkuPrice", "User", "UserAchievement",
    "NetworkManager",
    "OverlayManager",
    "RelationshipManager",
    "StorageManager",
    "StoreManager",
    "UserManager",
    "VoiceManager",
]



import ctypes
from typing import Callable

from . import sdk
from .enum import Result
from .event import bindEvents
from .exception import getException
from .model import Relationship


class RelationshipManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordRelationshipEvents,
            self._OnRefresh,
            self._OnRelationshipUpdate
        )

    def _OnRefresh(self, event_data):
        self.OnRefresh()

    def _OnRelationshipUpdate(self, event_data, relationship):
        self.OnRelationshipUpdate(Relationship(copy=relationship.contents))

    def Filter(self, filter: Callable[[Relationship], None]) -> None:
        """
        Filters a user's relationship list by a boolean condition.
        """
        def CFilter(filter_data, relationship):
            return bool(filter(Relationship(copy=relationship.contents)))

        CFilter = self._internal.filter.argtypes[-1](CFilter)

        self._internal.filter(self._internal, ctypes.c_void_p(), CFilter)

    def Get(self, userId: int) -> Relationship:
        """
        Get the relationship between the current user and a given user by id.
        """
        pointer = sdk.DiscordRelationship()
        result = Result(self._internal.get(self._internal, userId, pointer))
        if result != Result.Ok:
            raise getException(result)

        return Relationship(internal=pointer)

    def GetAt(self, index: int) -> Relationship:
        """
        Get the relationship at a given index when iterating over a list of relationships.
        """
        pointer = sdk.DiscordRelationship()
        result = Result(self._internal.get_at(self._internal, index, pointer))
        if result != Result.Ok:
            raise getException(result)

        return Relationship(internal=pointer)

    def Count(self) -> int:
        """
        Get the number of relationships that match your filter.
        """
        count = ctypes.c_int32()
        result = Result(self._internal.count(self._internal, count))
        if result != Result.Ok:
            raise getException(result)

        return count.value

    def OnRefresh(self) -> None:
        """
        Fires at initialization when Discord has cached a snapshot of the current status of all
        your relationships.
        """
        pass

    def OnRelationshipUpdate(self, relationship: Relationship) -> None:
        """
        Fires when a relationship in the filtered list changes, like an updated presence or user
        attribute.
        """
        pass



import ctypes
from typing import Callable

from . import sdk
from .enum import PremiumType, Result, UserFlag
from .event import bindEvents
from .exception import getException
from .model import User


class UserManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordUserEvents,
            self._OnCurrentUserUpdate
        )

    def _OnCurrentUserUpdate(self, event_data):
        self.OnCurrentUserUpdate()

    def GetCurrentUser(self) -> User:
        """
        Fetch information about the currently connected user account.
        """
        user = sdk.DiscordUser()
        result = Result(self._internal.get_current_user(self._internal, user))
        if result != Result.Ok:
            raise getException(result)

        return User(internal=user)

    def GetUser(self, userId: int, callback: Callable) -> None:
        """
        Get user information for a given id.

        Returns discord.enum.Result (int) and User via callback.
        """
        def CCallback(callback_data, result, user):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                callback(result, User(copy=user.contents))
            else:
                callback(result, None)

        CCallback = self._internal.get_user.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.get_user(self._internal, userId, ctypes.c_void_p(), CCallback)

    def GetCurrentUserPremiumType(self) -> PremiumType:
        """
        Get the PremiumType for the currently connected user.
        """
        premiumType = ctypes.c_int32()
        result = Result(self._internal.get_current_user_premium_type(self._internal, premiumType))
        if result != Result.Ok:
            raise getException(result)

        return PremiumType(premiumType.value)

    def CurrentUserHasFlag(self, flag: UserFlag) -> bool:
        """
        See whether or not the current user has a certain UserFlag on their account.
        """
        hasFlag = ctypes.c_bool()
        result = Result(self._internal.current_user_has_flag(self._internal, flag, hasFlag))
        if result != Result.Ok:
            raise getException(result)

        return hasFlag.value

    def OnCurrentUserUpdate(self) -> None:
        """
        Fires when the User struct of the currently connected user changes.
        """
        pass



import ctypes
from typing import Callable, Optional

from . import sdk
from .enum import Result
from .exception import getException
from .model import FileStat


class StorageManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = None

    def GetPath(self) -> str:
        """
        Returns the filepath to which Discord saves files if you were to use the SDK's storage
        manager.
        """
        path = sdk.DiscordPath()

        result = Result(self._internal.get_path(self._internal, path))
        if result != Result.Ok:
            raise getException(result)

        return path.value.decode("utf8")

    def Read(self, name: str) -> bytes:
        """
        Reads data synchronously from the game's allocated save file.
        """
        # we need the file stat for this one, as length-fixed buffers does not exist in python
        fileStat = self.Stat(name)
        fileSize = fileStat.Size

        name = ctypes.c_char_p(name.encode("utf8"))
        buffer = (ctypes.c_uint8 * fileSize)()
        read = ctypes.c_uint32()

        result = Result(self._internal.read(self._internal, name, buffer, len(buffer), read))
        if result != Result.Ok:
            raise getException(result)

        if read.value != fileSize:
            print("discord/storage.py: warning: attempting to read " +
                  str(fileSize) + " bytes, but read " + str(read.value))

        return bytes(buffer[:read.value])

    def ReadAsync(self, name: str, callback: Callable[[Result, Optional[bytes]], None]) -> None:
        """
        Reads data asynchronously from the game's allocated save file.

        Returns discord.enum.Result (int) and data (bytes) via callback.
        """
        def CCallback(callback_data, result, data, data_length):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                data = bytes(data[:data_length])
                callback(result, data)
            else:
                callback(result, None)

        CCallback = self._internal.read_async.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        name = ctypes.c_char_p(name.encode("utf8"))
        self._internal.read_async(self._internal, name, ctypes.c_void_p(), CCallback)

    def ReadAsyncPartial(
        self,
        name: str,
        offset: int,
        length: int,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Reads data asynchronously from the game's allocated save file, starting at a given offset
        and up to a given length.
        """
        def CCallback(callback_data, result, data, data_length):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                data = bytes(data[:data_length])
                callback(result, data)
            else:
                callback(result, None)

        CCallback = self._internal.read_async.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        name = ctypes.c_char_p(name.encode("utf8"))
        self._internal.read_async_partial(
            self._internal,
            name,
            offset,
            length,
            ctypes.c_void_p(),
            CCallback
        )

    def Write(self, name: str, data: bytes) -> None:
        """
        Writes data synchronously to disk, under the given key name.
        """
        name = ctypes.c_char_p(name.encode("utf8"))
        data = (ctypes.c_uint8 * len(data))(*data)

        result = Result(self._internal.write(self._internal, name, data, len(data)))
        if result != Result.Ok:
            raise getException(result)

    def WriteAsync(self, name: str, data: bytes, callback: Callable[[Result], None]) -> None:
        """
        Writes data asynchronously to disk under the given keyname.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.write_async.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        name = ctypes.c_char_p(name.encode("utf8"))
        data = (ctypes.c_uint8 * len(data))(*data)

        self._internal.write_async(
            self._internal,
            name,
            data,
            len(data),
            ctypes.c_void_p(),
            CCallback
        )

    def Delete(self, name: str) -> None:
        """
        Deletes written data for the given key name.
        """
        name = ctypes.c_char_p(name.encode("utf8"))

        result = Result(self._internal.delete_(self._internal, name))
        if result != Result.Ok:
            raise getException(result)

    def Exists(self, name: str) -> bool:
        """
        Checks if data exists for a given key name.
        """
        exists = ctypes.c_bool()
        name = ctypes.c_char_p(name.encode("utf8"))

        result = Result(self._internal.exists(self._internal, name, exists))
        if result != Result.Ok:
            raise getException(result)

        return exists.value

    def Stat(self, name: str) -> FileStat:
        """
        Returns file info for the given key name.
        """
        stat = sdk.DiscordFileStat()

        name = ctypes.c_char_p(name.encode("utf8"))
        result = Result(self._internal.stat(self._internal, name, stat))
        if result != Result.Ok:
            raise getException(result)

        return FileStat(internal=stat)

    def Count(self) -> int:
        """
        Returns the count of files, for iteration.
        """
        count = ctypes.c_int32()
        self._internal.count(self._internal, count)
        return count.value

    def StatAt(self, index: int) -> FileStat:
        """
        Returns file info for the given index when iterating over files.
        """
        stat = sdk.DiscordFileStat()

        result = Result(self._internal.stat_at(self._internal, index, stat))
        if result != Result.Ok:
            raise getException(result)

        return FileStat(internal=stat)



import ctypes
from typing import Callable, Optional

from . import sdk
from .enum import (
    LobbySearchCast, LobbySearchComparison, LobbySearchDistance, LobbyType,
    Result)
from .event import bindEvents
from .exception import getException
from .model import Lobby, User


class LobbyTransaction:
    def __init__(self, internal):
        self._internal = internal

    def SetType(self, type: LobbyType) -> None:
        """
        Marks a lobby as private or public.
        """
        result = Result(self._internal.set_type(self._internal, type))
        if result != Result.Ok:
            raise getException(result)

    def SetOwner(self, userId: int) -> None:
        """
        Sets a new owner for the lobby.
        """
        result = Result(self._internal.set_owner(self._internal, userId))
        if result != Result.Ok:
            raise getException(result)

    def SetCapacity(self, capacity: int) -> None:
        """
        Sets a new capacity for the lobby.
        """
        result = Result(self._internal.set_capacity(self._internal, capacity))
        if result != Result.Ok:
            raise getException(result)

    def SetMetadata(self, key: str, value: str) -> None:
        """
        Sets metadata value under a given key name for the lobby.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()
        metadataValue.value = value.encode("utf8")

        result = Result(self._internal.set_metadata(self._internal, metadataKey, metadataValue))
        if result != Result.Ok:
            raise getException(result)

    def DeleteMetadata(self, key: str) -> None:
        """
        Deletes the lobby metadata for a key.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        result = Result(self._internal.delete_metadata(self._internal, metadataKey))
        if result != Result.Ok:
            raise getException(result)

    def SetLocked(self, locked: bool) -> None:
        """
        Sets the lobby to locked or unlocked.
        """
        result = Result(self._internal.set_locked(self._internal, locked))
        if result != Result.Ok:
            raise getException(result)


class LobbyMemberTransaction:
    def __init__(self, internal):
        self._internal = internal

    def SetMetadata(self, key: str, value: str) -> None:
        """
        Sets metadata value under a given key name for the current user.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()
        metadataValue.value = value.encode("utf8")

        result = Result(self._internal.set_metadata(self._internal, metadataKey, metadataValue))
        if result != Result.Ok:
            raise getException(result)

    def DeleteMetadata(self, key: str) -> None:
        """
        Sets metadata value under a given key name for the current user.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        result = Result(self._internal.delete_metadata(self._internal, metadataKey))
        if result != Result.Ok:
            raise getException(result)


class LobbySearchQuery:
    def __init__(self, internal):
        self._internal = internal

    def Filter(
        self,
        key: str,
        comp: LobbySearchComparison,
        cast: LobbySearchCast,
        value: str
    ) -> None:
        """
        Filters lobbies based on metadata comparison.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()
        metadataValue.value = value.encode("utf8")

        result = Result(self._internal.filter(
            self._internal,
            metadataKey,
            comp,
            cast,
            metadataValue
        ))
        if result != Result.Ok:
            raise getException(result)

    def Sort(self, key: str, cast: LobbySearchCast, value: str) -> None:
        """
        Sorts the filtered lobbies based on "near-ness" to a given value.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()
        metadataValue.value = value.encode("utf8")

        result = Result(self._internal.sort(self._internal, metadataKey, cast, metadataValue))
        if result != Result.Ok:
            raise getException(result)

    def Limit(self, limit: int) -> None:
        """
        Limits the number of lobbies returned in a search.
        """
        result = Result(self._internal.limit(self._internal, limit))
        if result != Result.Ok:
            raise getException(result)

    def Distance(self, distance: LobbySearchDistance) -> None:
        """
        Filters lobby results to within certain regions relative to the user's location.
        """
        result = Result(self._internal.distance(self._internal, distance))
        if result != Result.Ok:
            raise getException(result)


class LobbyManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordLobbyEvents,
            self._OnLobbyUpdate,
            self._OnLobbyDelete,
            self._OnMemberConnect,
            self._OnMemberUpdate,
            self._OnMemberDisconnect,
            self._OnLobbyMessage,
            self._OnSpeaking,
            self._OnNetworkMessage
        )

    def _OnLobbyUpdate(self, event_data, lobby_id):
        self.OnLobbyUpdate(lobby_id)

    def _OnLobbyDelete(self, event_data, lobby_id, reason):
        self.OnLobbyDelete(lobby_id, reason)

    def _OnMemberConnect(self, event_data, lobby_id, user_id):
        self.OnMemberConnect(lobby_id, user_id)

    def _OnMemberUpdate(self, event_data, lobby_id, user_id):
        self.OnMemberUpdate(lobby_id, user_id)

    def _OnMemberDisconnect(self, event_data, lobby_id, user_id):
        self.OnMemberDisconnect(lobby_id, user_id)

    def _OnLobbyMessage(self, event_data, lobby_id, user_id, data, data_length):
        message = bytes(data[:data_length]).decode("utf8")
        self.OnLobbyMessage(lobby_id, user_id, message)

    def _OnSpeaking(self, event_data, lobby_id, user_id, speaking):
        self.OnSpeaking(lobby_id, user_id, speaking)

    def _OnNetworkMessage(self, event_data, lobby_id, user_id, channel_id, data, data_length):
        data = bytes(data[:data_length])
        self.OnNetworkMessage(lobby_id, user_id, channel_id, data)

    def GetLobbyCreateTransaction(self) -> LobbyTransaction:
        """
        Gets a Lobby transaction used for creating a new lobby
        """
        transaction = ctypes.POINTER(sdk.IDiscordLobbyTransaction)()
        result = Result(self._internal.get_lobby_create_transaction(self._internal, transaction))
        if result != Result.Ok:
            raise getException(result)

        return LobbyTransaction(internal=transaction.contents)

    def GetLobbyUpdateTransaction(self, lobbyId: int) -> LobbyTransaction:
        """
        Gets a lobby transaction used for updating an existing lobby.
        """
        transaction = ctypes.POINTER(sdk.IDiscordLobbyTransaction)()
        result = Result(self._internal.get_lobby_update_transaction(
            self._internal, lobbyId, transaction))
        if result != Result.Ok:
            raise getException(result)

        return LobbyTransaction(internal=transaction.contents)

    def GetMemberUpdateTransaction(self, lobbyId: int, userId: int) -> LobbyMemberTransaction:
        """
        Gets a new member transaction for a lobby member in a given lobby.
        """
        transaction = ctypes.POINTER(sdk.IDiscordLobbyMemberTransaction)()
        result = Result(self._internal.get_member_update_transaction(
            self._internal, lobbyId, userId, transaction))
        if result != Result.Ok:
            raise getException(result)

        return LobbyMemberTransaction(internal=transaction.contents)

    def CreateLobby(
        self,
        transaction:
        LobbyTransaction,
        callback: Callable[[Result, Optional[Lobby]], None]
    ) -> None:
        """
        Creates a lobby.

        Returns discord.enum.Result (int) and Lobby via callback.
        """
        def CCallback(callback_data, result, lobby):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                callback(result, Lobby(copy=lobby.contents))
            else:
                callback(result, None)

        CCallback = self._internal.create_lobby.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.create_lobby(
            self._internal,
            transaction._internal,
            ctypes.c_void_p(),
            CCallback
        )

    def UpdateLobby(
        self,
        lobbyId: int,
        transaction: LobbyTransaction,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Updates a lobby with data from the given transaction.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.update_lobby.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.update_lobby(
            self._internal,
            lobbyId,
            transaction._internal,
            ctypes.c_void_p(),
            CCallback
        )

    def DeleteLobby(self, lobbyId: int, callback: Callable[[Result], None]) -> None:
        """
        Deletes a given lobby.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.delete_lobby.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.delete_lobby(self._internal, lobbyId, ctypes.c_void_p(), CCallback)

    def ConnectLobby(
        self,
        lobbyId: int,
        lobbySecret: str,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Connects the current user to a given lobby.
        """
        def CCallback(callback_data, result, lobby):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                callback(result, Lobby(copy=lobby.contents))
            else:
                callback(result, None)

        CCallback = self._internal.connect_lobby.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        _lobbySecret = sdk.DiscordLobbySecret()
        _lobbySecret.value = lobbySecret.encode("utf8")

        self._internal.connect_lobby(
            self._internal,
            lobbyId,
            _lobbySecret,
            ctypes.c_void_p(),
            CCallback
        )

    def ConnectLobbyWithActivitySecret(
        self,
        activitySecret: str,
        callback: Callable[[Result, Optional[Lobby]], None]
    ) -> None:
        """
        Connects the current user to a lobby; requires the special activity secret from the lobby
        which is a concatenated lobbyId and secret.
        """
        def CCallback(callback_data, result, lobby):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                callback(result, Lobby(copy=lobby.contents))
            else:
                callback(result, None)

        CCallback = self._internal.connect_lobby_with_activity_secret.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        _activitySecret = sdk.DiscordLobbySecret()
        _activitySecret.value = activitySecret.encode("utf8")

        self._internal.connect_lobby_with_activity_secret(
            self._internal,
            _activitySecret,
            ctypes.c_void_p(),
            CCallback
        )

    def GetLobbyActivitySecret(self, lobbyId: int) -> str:
        """
        Gets the special activity secret for a given lobby.
        """
        lobbySecret = sdk.DiscordLobbySecret()

        result = self._internal.get_lobby_activity_secret(self._internal, lobbyId, lobbySecret)
        if result != result.Ok:
            raise getException(result)

        return lobbySecret.value.decode("utf8")

    def DisconnectLobby(self, lobbyId: int, callback: Callable[[Result], None]) -> None:
        """
        Disconnects the current user from a lobby.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.disconnect_lobby.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.disconnect_lobby(self._internal, lobbyId, ctypes.c_void_p(), CCallback)

    def GetLobby(self, lobbyId: int) -> Lobby:
        """
        Gets the lobby object for a given lobby id.
        """
        lobby = sdk.DiscordLobby()

        result = Result(self._internal.get_lobby(self._internal, lobbyId, lobby))
        if result != Result.Ok:
            raise getException(result)

        return Lobby(internal=lobby)

    def LobbyMetadataCount(self, lobbyId: int) -> int:
        """
        Returns the number of metadata key/value pairs on a given lobby.
        """
        count = sdk.c_int32()

        result = Result(self._internal.lobby_metadata_count(self._internal, lobbyId, count))
        if result != Result.Ok:
            raise getException(result)

        return count.value

    def GetLobbyMetadataKey(self, lobbyId: int, index: int) -> str:
        """
        Returns the key for the lobby metadata at the given index.
        """
        metadataKey = sdk.DiscordMetadataKey()

        result = Result(self._internal.get_lobby_metadata_key(
            self._internal,
            lobbyId,
            index,
            metadataKey
        ))
        if result != Result.Ok:
            raise getException(result)

        return metadataKey.value.decode("utf8")

    def GetLobbyMetadataValue(self, lobbyId: int, key: str) -> str:
        """
        Returns lobby metadata value for a given key and id.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()

        result = Result(self._internal.get_lobby_metadata_value(
            self._internal,
            lobbyId,
            metadataKey,
            metadataValue
        ))
        if result != Result.Ok:
            raise getException(result)

        return metadataValue.value.decode("utf8")

    def MemberCount(self, lobbyId: int) -> int:
        """
        Get the number of members in a lobby.
        """
        count = sdk.c_int32()

        result = Result(self._internal.member_count(self._internal, lobbyId, count))
        if result != Result.Ok:
            raise getException(result)

        return count.value

    def GetMemberUserId(self, lobbyId: int, index: int) -> int:
        """
        Gets the user id of the lobby member at the given index.
        """
        userId = sdk.DiscordUserId()

        result = Result(self._internal.get_member_user_id(self._internal, lobbyId, index, userId))
        if result != Result.Ok:
            raise getException(result)

        return userId.value

    def GetMemberUser(self, lobbyId: int, userId: int) -> User:
        """
        Gets the user object for a given user id.
        """
        user = sdk.DiscordUser()

        result = Result(self._internal.get_member_user(self._internal, lobbyId, userId, user))
        if result != Result.Ok:
            raise getException(result)

        return User(internal=user)

    def MemberMetadataCount(self, lobbyId: int, userId: int) -> int:
        """
        Gets the number of metadata key/value pairs for the given lobby member.
        """
        count = sdk.c_int32()

        result = Result(self._internal.member_metadata_count(
            self._internal,
            lobbyId,
            userId,
            count
        ))
        if result != Result.Ok:
            raise getException(result)

        return count.value

    def GetMemberMetadataKey(self, lobbyId: int, userId: int, index: int) -> str:
        """
        Gets the key for the lobby metadata at the given index on a lobby member.
        """
        metadataKey = sdk.DiscordMetadataKey()

        result = Result(self._internal.get_member_metadata_key(
            self._internal,
            lobbyId,
            userId,
            index,
            metadataKey
        ))
        if result != Result.Ok:
            raise getException(result)

        return metadataKey.value.decode("utf8")

    def GetMemberMetadataValue(self, lobbyId: int, userId: int, key: str) -> str:
        """
        Returns user metadata for a given key.
        """
        metadataKey = sdk.DiscordMetadataKey()
        metadataKey.value = key.encode("utf8")

        metadataValue = sdk.DiscordMetadataValue()

        result = Result(self._internal.get_member_metadata_value(
            self._internal,
            lobbyId,
            userId,
            metadataKey,
            metadataValue
        ))
        if result != Result.Ok:
            raise getException(result)

        return metadataValue.value.decode("utf8")

    def UpdateMember(
        self,
        lobbyId: int,
        userId: int,
        transaction: LobbyMemberTransaction,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Updates lobby member info for a given member of the lobby.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.update_member.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.update_member(
            self._internal,
            lobbyId,
            userId,
            transaction._internal,
            ctypes.c_void_p(),
            CCallback
        )

    def SendLobbyMessage(self, lobbyId: int, data: str, callback: Callable[[Result], None]) -> None:
        """
        Sends a message to the lobby on behalf of the current user.

        Returns Discord.result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.send_lobby_message.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        data = data.encode("utf8")
        data = (ctypes.c_uint8 * len(data))(*data)
        self._internal.send_lobby_message(
            self._internal, lobbyId, data, len(data), ctypes.c_void_p(), CCallback)

    def GetSearchQuery(self) -> LobbySearchQuery:
        """
        Creates a search object to search available lobbies.
        """
        search_query = (ctypes.POINTER(sdk.IDiscordLobbySearchQuery))()
        result = Result(self._internal.get_search_query(self._internal, ctypes.byref(search_query)))
        if result != Result.Ok:
            raise getException(result)

        return LobbySearchQuery(internal=search_query.contents)

    def Search(self, search: LobbySearchQuery, callback: Callable[[Result], None]) -> None:
        """
        Searches available lobbies based on the search criteria chosen in the LobbySearchQuery
        member functions.

        Lobbies that meet the criteria are then globally filtered, and can be accessed via
        iteration with LobbyCount() and GetLobbyId(). The callback fires when the list of lobbies
        is stable and ready for iteration.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.search.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.search(self._internal, search._internal, ctypes.c_void_p(), CCallback)

    def LobbyCount(self) -> int:
        """
        Get the number of lobbies that match the search.
        """
        count = sdk.c_int32()
        self._internal.lobby_count(self._internal, count)
        return count.value

    def GetLobbyId(self, index: int) -> int:
        """
        Returns the id for the lobby at the given index.
        """

        lobbyId = sdk.DiscordLobbyId()

        result = Result(self._internal.get_lobby_id(self._internal, index, lobbyId))
        if result != Result.Ok:
            raise getException(result)

        return lobbyId.value

    def ConnectVoice(self, lobbyId: int, callback: Callable[[Result], None]) -> None:
        """
        Connects to the voice channel of the current lobby.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.connect_voice.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.connect_voice(self._internal, lobbyId, ctypes.c_void_p(), CCallback)

    def DisconnectVoice(self, lobbyId: int, callback: Callable[[Result], None]) -> None:
        """
        Disconnects from the voice channel of a given lobby.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.disconnect_voice.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.disconnect_voice(self._internal, lobbyId, ctypes.c_void_p(), CCallback)

    def OnLobbyUpdate(self, lobbyId: int) -> None:
        """
        Fires when a lobby is updated.
        """
        pass

    def OnLobbyDelete(self, lobbyId: int, reason: str) -> None:
        """
        Fired when a lobby is deleted.
        """
        pass

    def OnMemberConnect(self, lobbyId: int, userId: int) -> None:
        """
        Fires when a new member joins the lobby.
        """
        pass

    def OnMemberUpdate(self, lobbyId: int, userId: int) -> None:
        """
        Fires when data for a lobby member is updated.
        """
        pass

    def OnMemberDisconnect(self, lobbyId: int, userId: int) -> None:
        """
        Fires when a member leaves the lobby.
        """
        pass

    def OnLobbyMessage(self, lobbyId: int, userId: int, message: str) -> None:
        """
        Fires when a message is sent to the lobby.
        """
        pass

    def OnSpeaking(self, lobbyId: int, userId: int, speaking: bool) -> None:
        """
        Fires when a user connected to voice starts or stops speaking.
        """
        pass

    def ConnectNetwork(self, lobbyId: int) -> None:
        """
        Connects to the networking layer for the given lobby ID.
        """
        result = Result(self._internal.connect_network(self._internal, lobbyId))
        if result != Result.Ok:
            raise getException(result)

    def DisconnectNetwork(self, lobbyId: int) -> None:
        """
        Disconnects from the networking layer for the given lobby ID.
        """
        result = Result(self._internal.disconnect_network(self._internal, lobbyId))
        if result != Result.Ok:
            raise getException(result)

    def FlushNetwork(self) -> None:
        """
        Flushes the network. Call this when you're done sending messages.
        """
        result = Result(self._internal.flush_network(self._internal))
        if result != Result.Ok:
            raise getException(result)

    def OpenNetworkChannel(self, lobbyId: int, channelId: int, reliable: bool) -> None:
        """
        Opens a network channel to all users in a lobby on the given channel number. No need to
        iterate over everyone!
        """
        result = Result(self._internal.open_network_channel(
            self._internal, lobbyId, channelId, reliable))
        if result != Result.Ok:
            raise getException(result)

    def SendNetworkMessage(self, lobbyId: int, userId: int, channelId: int, data: bytes) -> None:
        """
        Sends a network message to the given user ID that is a member of the given lobby ID over
        the given channel ID.
        """
        data = (ctypes.c_uint8 * len(data))(*data)
        result = Result(self._internal.send_network_message(
            self._internal, lobbyId, userId, channelId, data, len(data)))
        if result != Result.Ok:
            raise getException(result)

    def OnNetworkMessage(self, lobbyId: int, userId: int, channelId: int, data: bytes) -> None:
        """
        Fires when the user receives a message from the lobby's networking layer.
        """
        pass



import ctypes
from typing import Callable

from . import sdk
from .achievement import AchievementManager
from .activity import ActivityManager
from .application import ApplicationManager
from .enum import CreateFlags, LogLevel, Result
from .exception import getException
from .image import ImageManager
from .lobby import LobbyManager
from .network import NetworkManager
from .overlay import OverlayManager
from .relationship import RelationshipManager
from .storage import StorageManager
from .store import StoreManager
from .user import UserManager
from .voice import VoiceManager


class Discord:
    def __init__(self, clientId: int, flags: CreateFlags):
        self.core = None

        self._activityManager = ActivityManager()
        self._relationshipManager = RelationshipManager()
        self._imageManager = ImageManager()
        self._userManager = UserManager()
        self._lobbyManager = LobbyManager()
        self._networkManager = NetworkManager()
        self._overlayManager = OverlayManager()
        self._applicationManager = ApplicationManager()
        self._storageManager = StorageManager()
        self._storeManager = StoreManager()
        self._voiceManager = VoiceManager()
        self._achievementManager = AchievementManager()

        self._garbage = []

        version = sdk.DiscordVersion(2)

        params = sdk.DiscordCreateParams()
        params.client_id = clientId
        params.flags = flags

        sdk.DiscordCreateParamsSetDefault(params)
        params.activity_events = self._activityManager._events
        params.relationship_events = self._relationshipManager._events
        params.image_events = self._imageManager._events
        params.user_events = self._userManager._events
        params.lobby_events = self._lobbyManager._events
        params.network_events = self._networkManager._events
        params.overlay_events = self._overlayManager._events
        params.application_events = self._applicationManager._events
        params.storage_events = self._storageManager._events
        params.store_events = self._storeManager._events
        params.voice_events = self._voiceManager._events
        params.achievement_events = self._achievementManager._events

        pointer = ctypes.POINTER(sdk.IDiscordCore)()

        result = Result(sdk.DiscordCreate(version, ctypes.pointer(params), ctypes.pointer(pointer)))
        if result != Result.Ok:
            raise getException(result)

        self.core = pointer.contents

    def __del__(self):
        if self.core:
            self.core.destroy(self.core)
            self.core = None

    def SetLogHook(self, min_level: LogLevel, hook: Callable[[LogLevel, str], None]) -> None:
        """
        Registers a logging callback function with the minimum level of message to receive.
        """
        def CHook(hook_data, level, message):
            level = LogLevel(level)
            hook(level, message.decode("utf8"))

        CHook = self.core.set_log_hook.argtypes[-1](CHook)
        self._garbage.append(CHook)

        self.core.set_log_hook(self.core, min_level.value, ctypes.c_void_p(), CHook)

    def RunCallbacks(self) -> None:
        """
        Runs all pending SDK callbacks.
        """
        result = Result(self.core.run_callbacks(self.core))
        if result != Result.Ok:
            raise getException(result)

    def GetActivityManager(self) -> ActivityManager:
        """
        Fetches an instance of the manager for interfacing with activies in the SDK.
        """
        if not self._activityManager._internal:
            self._activityManager._internal = self.core.get_activity_manager(self.core).contents

        return self._activityManager

    def GetRelationshipManager(self) -> RelationshipManager:
        """
        Fetches an instance of the manager for interfacing with relationships in the SDK.
        """
        if not self._relationshipManager._internal:
            self._relationshipManager._internal = self.core.get_relationship_manager(self.core).contents  # noqa: E501

        return self._relationshipManager

    def GetImageManager(self) -> ImageManager:
        """
        Fetches an instance of the manager for interfacing with images in the SDK.
        """
        if not self._imageManager._internal:
            self._imageManager._internal = self.core.get_image_manager(self.core).contents

        return self._imageManager

    def GetUserManager(self) -> UserManager:
        """
        Fetches an instance of the manager for interfacing with users in the SDK.
        """
        if not self._userManager._internal:
            self._userManager._internal = self.core.get_user_manager(self.core).contents

        return self._userManager

    def GetLobbyManager(self) -> LobbyManager:
        """
        Fetches an instance of the manager for interfacing with lobbies in the SDK.
        """
        if not self._lobbyManager._internal:
            self._lobbyManager._internal = self.core.get_lobby_manager(self.core).contents

        return self._lobbyManager

    def GetNetworkManager(self) -> NetworkManager:
        """
        Fetches an instance of the manager for interfacing with networking in the SDK.
        """
        if not self._networkManager._internal:
            self._networkManager._internal = self.core.get_network_manager(self.core).contents

        return self._networkManager

    def GetOverlayManager(self) -> OverlayManager:
        """
        Fetches an instance of the manager for interfacing with the overlay in the SDK.
        """
        if not self._overlayManager._internal:
            self._overlayManager._internal = self.core.get_overlay_manager(self.core).contents

        return self._overlayManager

    def GetApplicationManager(self) -> ApplicationManager:
        """
        Fetches an instance of the manager for interfacing with applications in the SDK.
        """
        if not self._applicationManager._internal:
            self._applicationManager._internal = self.core.get_application_manager(self.core).contents  # noqa: E501

        return self._applicationManager

    def GetStorageManager(self) -> StorageManager:
        """
        Fetches an instance of the manager for interfacing with storage in the SDK.
        """
        if not self._storageManager._internal:
            self._storageManager._internal = self.core.get_storage_manager(self.core).contents

        return self._storageManager

    def GetStoreManager(self) -> StoreManager:
        """
        Fetches an instance of the manager for interfacing with SKUs and Entitlements in the SDK.
        """
        if not self._storeManager._internal:
            self._storeManager._internal = self.core.get_store_manager(self.core).contents

        return self._storeManager

    def GetVoiceManager(self) -> VoiceManager:
        """
        Fetches an instance of the manager for interfacing with voice chat in the SDK.
        """
        if not self._voiceManager._internal:
            self._voiceManager._internal = self.core.get_voice_manager(self.core).contents

        return self._voiceManager

    def GetAchievementManager(self) -> AchievementManager:
        """
        Fetches an instance of the manager for interfacing with achievements in the SDK.
        """
        if not self._achievementManager._internal:
            self._achievementManager._internal = self.core.get_achievement_manager(self.core).contents  # noqa: E501

        return self._achievementManager



import sys

if sys.version_info >= (3, 6):
    from enum import IntEnum, IntFlag
else:
    from enum import IntEnum, IntEnum as IntFlag


class Result(IntEnum):
    Ok = 0
    ServiceUnavailable = 1
    InvalidVersion = 2
    LockFailed = 3
    InternalError = 4
    InvalidPayload = 5
    InvalidCommand = 6
    InvalidPermissions = 7
    NotFetched = 8
    NotFound = 9
    Conflict = 10
    InvalidSecret = 11
    InvalidJoinSecret = 12
    NoEligibleActivity = 13
    InvalidInvite = 14
    NotAuthenticated = 15
    InvalidAccessToken = 16
    ApplicationMismatch = 17
    InvalidDataUrl = 18
    InvalidBase64 = 19
    NotFiltered = 20
    LobbyFull = 21
    InvalidLobbySecret = 22
    InvalidFilename = 23
    InvalidFileSize = 24
    InvalidEntitlement = 25
    NotInstalled = 26
    NotRunning = 27
    InsufficientBuffer = 28
    PurchaseCanceled = 29
    InvalidGuild = 30
    InvalidEvent = 31
    InvalidChannel = 32
    InvalidOrigin = 33
    RateLimited = 34
    OAuth2Error = 35
    SelectChannelTimeout = 36
    GetGuildTimeout = 37
    SelectVoiceForceRequired = 38
    CaptureShortcutAlreadyListening = 39
    UnauthorizedForAchievement = 40
    InvalidGiftCode = 41
    PurchaseError = 42
    TransactionAborted = 43
    # DrawingInitFailed = 44


class LogLevel(IntEnum):
    Error = 0
    Warning = 1
    Info = 2
    Debug = 3


class CreateFlags(IntFlag):
    Default = 0
    NoRequireDiscord = 1


class UserFlag(IntFlag):
    Partner = 2
    HypeSquadEvents = 4
    HypeSquadHouse1 = 64
    HypeSquadHouse2 = 128
    HypeSquadHouse3 = 256


class PremiumType(IntEnum):
    None_ = 0
    Tier1 = 1
    Tier2 = 2


class ActivityType(IntEnum):
    Playing = 0
    Streaming = 1
    Listening = 2
    Custom = 4


class ActivityJoinRequestReply(IntEnum):
    No = 0
    Yes = 1
    Ignore = 2


class ActivityActionType(IntEnum):
    Join = 1
    Spectate = 2


class RelationshipType(IntEnum):
    None_ = 0
    Friend = 1
    Blocked = 2
    PendingIncoming = 3
    PendingOutgoing = 4
    Implicit = 5


class Status(IntEnum):
    Offline = 0
    Online = 1
    Idle = 2
    DoNotDisturb = 3


class ImageType(IntEnum):
    User = 0


class LobbyType(IntEnum):
    Private = 1
    Public = 2


class LobbySearchComparison(IntEnum):
    LessThanOrEqual = -2
    LessThan = -1
    Equal = 0
    GreaterThan = 1
    GreaterThanOrEqual = 2
    NotEqual = 3


class LobbySearchCast(IntEnum):
    String = 1
    Number = 2


class LobbySearchDistance(IntEnum):
    Local = 0
    Default = 1
    Extended = 2
    Global = 3


class InputModeType(IntEnum):
    VoiceActivity = 0
    PushToTalk = 1


class SkuType(IntEnum):
    Application = 1
    DLC = 2
    Consumable = 3
    Bundle = 4


class EntitlementType(IntEnum):
    Purchase = 1
    PremiumSubscription = 2
    DeveloperGift = 3
    TestModePurchase = 4
    FreePurchase = 5
    UserGift = 6
    PremiumPurchase = 7



import ctypes

from . import sdk
from .enum import Result
from .event import bindEvents
from .exception import getException


class NetworkManager:
    def __init__(self):
        self._internal = None
        self._events = bindEvents(
            sdk.IDiscordNetworkEvents,
            self._OnMessage,
            self._OnRouteUpdate
        )

    def _OnMessage(self, event_data, peer_id, channel_id, data, data_length):
        data = bytes(data[:data_length])
        self.OnMessage(peer_id, channel_id, data)

    def _OnRouteUpdate(self, event_data, route_data):
        self.OnRouteUpdate(route_data.decode("utf8"))

    def GetPeerId(self) -> int:
        """
        Get the networking peer ID for the current user, allowing other users to send packets to
        them.
        """
        peerId = sdk.DiscordNetworkPeerId()
        self._internal.get_peer_id(self._internal, peerId)
        return peerId.value

    def Flush(self) -> None:
        """
        Flushes the network
        """
        result = Result(self._internal.flush(self._internal))
        if result != Result.Ok:
            raise getException(result)

    def OpenChannel(self, peerId: int, channelId: int, reliable: bool) -> None:
        """
        Opens a channel to a user with their given peer ID on the given channel number.
        """
        result = Result(self._internal.open_channel(self._internal, peerId, channelId, reliable))
        if result != Result.Ok:
            raise getException(result)

    def OpenPeer(self, peerId: int, route: str) -> None:
        """
        Opens a network connection to another Discord user.
        """
        route_data = ctypes.create_string_buffer(route.encode("utf8"))
        result = Result(self._internal.open_peer(self._internal, peerId, route_data))
        if result != Result.Ok:
            raise getException(result)

    def UpdatePeer(self, peerId: int, route: str) -> None:
        """
        Updates the network connection to another Discord user.
        """
        route_data = ctypes.create_string_buffer(route.encode("utf8"))
        result = Result(self._internal.update_peer(self._internal, peerId, route_data))
        if result != Result.Ok:
            raise getException(result)

    def SendMessage(self, peerId: int, channelId: int, data: bytes) -> None:
        """
        Sends data to a given peer ID through the given channel.
        """
        data = (ctypes.c_uint8 * len(data))(*data)
        result = Result(self._internal.send_message(
            self._internal, peerId, channelId, data, len(data)))
        if result != Result.Ok:
            raise getException(result)

    def CloseChannel(self, peerId: int, channelId: int) -> None:
        """
        Close the connection to a given user by peerId on the given channel.
        """
        result = Result(self._internal.close_channel(self._internal, peerId, channelId))
        if result != Result.Ok:
            raise getException(result)

    def ClosePeer(self, peerId: int) -> None:
        """
        Disconnects the network session to another Discord user.
        """
        result = Result(self._internal.close_peer(self._internal, peerId))
        if result != Result.Ok:
            raise getException(result)

    def OnMessage(self, peerId: int, channelId: int, data: bytes) -> None:
        """
        Fires when you receive data from another user.
        """
        pass

    def OnRouteUpdate(self, route: str) -> None:
        """
        Fires when your networking route has changed.
        """
        pass



import ctypes
from typing import Callable

from . import sdk
from .enum import Result
from .event import bindEvents
from .exception import getException
from .model import Entitlement, Sku


class StoreManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordStoreEvents,
            self._OnEntitlementCreate,
            self._OnEntitlementDelete
        )

    def _OnEntitlementCreate(self, event_data, entitlement):
        self.OnEntitlementCreate(Entitlement(copy=entitlement))

    def _OnEntitlementDelete(self, event_data, entitlement):
        self.OnEntitlementDelete(Entitlement(copy=entitlement))

    def FetchSkus(self, callback: Callable[[Result], None]) -> None:
        """
        Fetches the list of SKUs for the connected application, readying them for iteration.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.fetch_skus.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.fetch_skus(self._internal, ctypes.c_void_p(), CCallback)

    def CountSkus(self) -> int:
        """
        Get the number of SKUs readied by FetchSkus().
        """
        count = ctypes.c_int32()
        self._internal.count_skus(self._internal, count)
        return count.value

    def GetSku(self, skuId: int) -> Sku:
        """
        Gets a SKU by its ID.
        """
        sku = sdk.DiscordSku()

        result = Result(self._internal.get_sku(skuId, sku))
        if result != Result.Ok:
            raise getException(result)

        return Sku(internal=sku)

    def GetSkuAt(self, index: int) -> Sku:
        """
        Gets a SKU by index when iterating over SKUs.
        """
        sku = sdk.DiscordSku()

        result = Result(self._internal.get_sku_at(index, sku))
        if result != Result.Ok:
            raise getException(result)

        return Sku(internal=sku)

    def FetchEntitlements(self, callback: Callable[[Result], None]) -> None:
        """
        Fetches a list of entitlements to which the user is entitled.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.fetch_entitlements.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.fetch_entitlements(self._internal, ctypes.c_void_p(), CCallback)

    def CountEntitlements(self) -> int:
        """
        Get the number of entitlements readied by FetchEntitlements().
        """
        count = ctypes.c_int32()
        self._internal.count_entitlements(self._internal, count)
        return count.value

    def GetEntitlement(self, entitlementId: int) -> Entitlement:
        """
        Gets an entitlement by its id.
        """
        entitlement = sdk.DiscordEntitlement()

        result = Result(self._internal.get_entitlement(entitlementId, entitlement))
        if result != Result.Ok:
            raise getException(result)

        return Entitlement(internal=Sku)

    def GetEntitlementAt(self, index: int) -> Entitlement:
        """
        Gets an entitlement by index when iterating over a user's entitlements.
        """
        entitlement = sdk.DiscordEntitlement()

        result = Result(self._internal.get_entitlement_at(index, entitlement))
        if result != Result.Ok:
            raise getException(result)

        return Entitlement(internal=Sku)

    def HasSkuEntitlement(self, skuId: int) -> bool:
        """
        Returns whether or not the user is entitled to the given SKU ID.
        """
        has_entitlement = ctypes.c_bool()

        result = Result(self._internal.has_sku_entitlement(skuId, has_entitlement))
        if result != Result.Ok:
            raise getException(result)

        return has_entitlement.value

    def StartPurchase(self, skuId: int, callback: Callable[[Result], None]) -> None:
        """
        Opens the overlay to begin the in-app purchase dialogue for the given SKU ID.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.start_purchase.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.start_purchase(self._internal, skuId, ctypes.c_void_p(), CCallback)

    def OnEntitlementCreate(self, entitlement: Entitlement) -> None:
        """
        Fires when the connected user receives a new entitlement, either through purchase or
        through a developer grant.
        """
        pass

    def OnEntitlementDelete(self, entitlement: Entitlement) -> None:
        """
        Fires when the connected user loses an entitlement, either by expiration, revocation, or
        consumption in the case of consumable entitlements.
        """
        pass



import ctypes
from typing import Callable, Optional

from . import sdk
from .enum import Result
from .exception import getException
from .model import ImageDimensions, ImageHandle


class ImageManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = None

    def Fetch(
        self,
        handle: ImageHandle,
        refresh: bool,
        callback: Callable[[Result, Optional[ImageHandle]], None]
    ) -> None:
        """
        Prepares an image to later retrieve data about it.

        Returns discord.enum.Result (int) and ImageHandle via callback.
        """
        def CCallback(callback_data, result, handle):
            self._garbage.remove(CCallback)
            result = Result(result)
            if result == Result.Ok:
                callback(result, ImageHandle(internal=handle))
            else:
                callback(result, None)

        CCallback = self._internal.fetch.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.fetch(
            self._internal,
            handle._internal,
            refresh,
            ctypes.c_void_p(),
            CCallback
        )

    def GetDimensions(self, handle: ImageHandle) -> ImageDimensions:
        """
        Gets the dimension for the given user's avatar's source image
        """
        dimensions = sdk.DiscordImageDimensions()
        result = Result(self._internal.get_dimensions(self._internal, handle._internal, dimensions))
        if result != Result.Ok:
            raise getException(result)

        return ImageDimensions(internal=dimensions)

    def GetData(self, handle: ImageHandle) -> bytes:
        """
        Gets the image data for a given user's avatar.
        """
        dimensions = self.GetDimensions(handle)
        buffer = (ctypes.c_uint8 * (dimensions.Width * dimensions.Height * 4))()

        result = Result(self._internal.get_data(
            self._internal, handle._internal, buffer, len(buffer)))
        if result != Result.Ok:
            raise getException(result)

        return bytes(buffer)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="discordsdk",
    version="0.1.0.post1",
    author="DoAltPlusF4",
    author_email="doaltplusf4@gmail.com",
    description="Python wrapper around Discord's Game SDK library.",
    license="LICENSE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DoAltPlusF4/discord-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



import ctypes
from typing import Callable, Optional

from . import sdk
from .enum import Result
from .model import OAuth2Token


class SignedAppTicket:
    def __init__(self):
        self.application_id = None
        self.user = None
        self.entitlements = None
        self.timestamp = None


class ApplicationManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = None

    def GetCurrentLocale(self) -> str:
        """
        Get the locale the current user has Discord set to.
        """
        locale = sdk.DiscordLocale()
        self._internal.get_current_locale(self._internal, locale)
        return locale.value.decode("utf8")

    def GetCurrentBranch(self) -> str:
        """
        Get the name of pushed branch on which the game is running.
        """
        branch = sdk.DiscordBranch()
        self._internal.get_current_branch(self._internal, branch)
        return branch.value.decode("utf8")

    def GetOAuth2Token(self, callback: Callable[[Result, Optional[OAuth2Token]], None]) -> None:
        """
        Retrieve an oauth2 bearer token for the current user.

        Returns discord.enum.Result (int) and OAuth2Token (str) via callback.
        """
        def CCallback(callback_data, result, oauth2_token):
            self._garbage.remove(CCallback)
            if result == Result.Ok:
                callback(result, OAuth2Token(copy=oauth2_token.contents))
            else:
                callback(result, None)

        CCallback = self._internal.get_oauth2_token.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.get_oauth2_token(self._internal, ctypes.c_void_p(), CCallback)

    def ValidateOrExit(self, callback: Callable[[Result], None]) -> None:
        """
        Checks if the current user has the entitlement to run this game.

        Returns discord.enum.Result (int) via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            callback(result)

        CCallback = self._internal.validate_or_exit.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.validate_or_exit(self._internal, ctypes.c_void_p(), CCallback)

    def GetTicket(self, callback: Callable[[Result, Optional[str]], None]) -> None:
        """
        Get the signed app ticket for the current user.

        Returns discord.Enum.Result (int) and str via callback.
        """
        def CCallback(callback_data, result, data):
            self._garbage.remove(CCallback)
            if result == Result.Ok:
                callback(result, data.contents.value.decode("utf8"))
            else:
                callback(result, None)

        CCallback = self._internal.get_ticket.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.get_ticket(self._internal, ctypes.c_void_p(), CCallback)



import ctypes
from typing import Callable

from . import sdk
from .enum import Result
from .event import bindEvents
from .exception import getException
from .model import UserAchievement


class AchievementManager:
    def __init__(self):
        self._internal = None
        self._garbage = []
        self._events = bindEvents(
            sdk.IDiscordAchievementEvents,
            self._OnUserAchievementUpdate
        )

    def _OnUserAchievementUpdate(self, event_data, user_achievement):
        self.OnUserAchievementUpdate(UserAchievement(copy=user_achievement.contents))

    def SetUserAchievement(
        self,
        achievementId: int,
        percentComplete: int,
        callback: Callable[[Result], None]
    ) -> None:
        """
        Updates the current user's status for a given achievement.

        Returns discord.enum.Result via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.set_user_achievement.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.set_user_achievement(
            self._internal,
            achievementId,
            percentComplete,
            ctypes.c_void_p(),
            CCallback
        )

    def FetchUserAchievements(self, callback: Callable[[Result], None]) -> None:
        """
        Loads a stable list of the current user's achievements to iterate over.

        Returns discord.enum.Result via callback.
        """
        def CCallback(callback_data, result):
            self._garbage.remove(CCallback)
            result = Result(result)
            callback(result)

        CCallback = self._internal.fetch_user_achievements.argtypes[-1](CCallback)
        self._garbage.append(CCallback)  # prevent it from being garbage collected

        self._internal.fetch_user_achievements(self._internal, ctypes.c_void_p(), CCallback)

    def CountUserAchievements(self) -> int:
        """
        Counts the list of a user's achievements for iteration.
        """
        count = ctypes.c_int32()
        self._internal.count_user_achievements(self._internal, count)
        return count.value

    def GetUserAchievementAt(self, index: int) -> UserAchievement:
        """
        Gets the user's achievement at a given index of their list of achievements.
        """
        achievement = sdk.DiscordUserAchievement()
        result = Result(self._internal.get_user_achievement_at(
            self._internal,
            index,
            achievement
        ))
        if result != Result.Ok:
            raise getException(result)

        return UserAchievement(internal=achievement)

    def GetUserAchievement(self, achievementId: int) -> None:
        """
        Gets the user achievement for the given achievement id.
        """
        achievement = sdk.DiscordUserAchievement()
        result = Result(self._internal.get_user_achievement(
            self._internal,
            achievementId,
            achievement
        ))
        if result != Result.Ok:
            raise getException(result)

        return UserAchievement(internal=achievement)

    def OnUserAchievementUpdate(self, achievement: UserAchievement) -> None:
        """
        Fires when an achievement is updated for the currently connected user
        """
        pass



import os.path
import ctypes

dll = ctypes.CDLL(os.path.abspath("lib/discord_game_sdk"))
DiscordCreate = dll.DiscordCreate

DiscordClientId = ctypes.c_int64
DiscordVersion = ctypes.c_int32
DiscordSnowflake = ctypes.c_int64
DiscordTimestamp = ctypes.c_int64
DiscordUserId = DiscordSnowflake
DiscordLocale = ctypes.c_char * 128
DiscordBranch = ctypes.c_char * 4096
DiscordLobbyId = DiscordSnowflake
DiscordLobbySecret = ctypes.c_char * 128
DiscordMetadataKey = ctypes.c_char * 256
DiscordMetadataValue = ctypes.c_char * 4096
DiscordNetworkPeerId = ctypes.c_uint64
DiscordNetworkChannelId = ctypes.c_uint8
DiscordPath = ctypes.c_char * 4096
DiscordDateTime = ctypes.c_char * 64


class DiscordUser(ctypes.Structure):
    _fields_ = [
        ("id", DiscordUserId),
        ("username", ctypes.c_char * 256),
        ("discriminator", ctypes.c_char * 8),
        ("avatar", ctypes.c_char * 128),
        ("bot", ctypes.c_bool),
    ]


class DiscordOAuth2Token(ctypes.Structure):
    _fields_ = [
        ("access_token", ctypes.c_char * 128),
        ("scopes", ctypes.c_char * 1024),
        ("expires", DiscordTimestamp),
    ]


class DiscordImageHandle(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int32),
        ("id", ctypes.c_int64),
        ("size", ctypes.c_uint32),
    ]


class DiscordImageDimensions(ctypes.Structure):
    _fields_ = [
        ("width", ctypes.c_uint32),
        ("height", ctypes.c_uint32),
    ]


class DiscordActivityTimestamps(ctypes.Structure):
    _fields_ = [
        ("start", DiscordTimestamp),
        ("end", DiscordTimestamp),
    ]


class DiscordActivityAssets(ctypes.Structure):
    _fields_ = [
        ("large_image", ctypes.c_char * 128),
        ("large_text", ctypes.c_char * 128),
        ("small_image", ctypes.c_char * 128),
        ("small_text", ctypes.c_char * 128),
    ]


class DiscordPartySize(ctypes.Structure):
    _fields_ = [
        ("current_size", ctypes.c_int32),
        ("max_size", ctypes.c_int32),
    ]


class DiscordActivityParty(ctypes.Structure):
    _fields_ = [
        ("id", ctypes.c_char * 128),
        ("size", DiscordPartySize),
    ]


class DiscordActivitySecrets(ctypes.Structure):
    _fields_ = [
        ("match", ctypes.c_char * 128),
        ("join", ctypes.c_char * 128),
        ("spectate", ctypes.c_char * 128),
    ]


class DiscordActivity(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int32),
        ("application_id", ctypes.c_uint64),
        ("name", ctypes.c_char * 128),
        ("state", ctypes.c_char * 128),
        ("details", ctypes.c_char * 128),
        ("timestamps", DiscordActivityTimestamps),
        ("assets", DiscordActivityAssets),
        ("party", DiscordActivityParty),
        ("secrets", DiscordActivitySecrets),
        ("instance", ctypes.c_bool),
    ]


class DiscordPresence(ctypes.Structure):
    _fields_ = [
        ("status", ctypes.c_int32),
        ("activity", DiscordActivity),
    ]


class DiscordRelationship(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int32),
        ("user", DiscordUser),
        ("presence", DiscordPresence),
    ]


class DiscordLobby(ctypes.Structure):
    _fields_ = [
        ("id", DiscordLobbyId),
        ("type", ctypes.c_int32),
        ("owner_id", DiscordUserId),
        ("secret", DiscordLobbySecret),
        ("capacity", ctypes.c_uint32),
        ("locked", ctypes.c_bool),
    ]


# SDK VERSION 2.5.7+ STUFF
"""
class DiscordImeUnderline(ctypes.Structure):
    _fields_ = [
        ("from", ctypes.c_int32),
        ("to", ctypes.c_int32),
        ("color", ctypes.c_int32),
        ("background_color", ctypes.c_uint32),
        ("thick", ctypes.c_bool),
    ]

class DiscordRect(ctypes.Structure):
    _fields_ = [
        ("left", ctypes.c_int32),
        ("top", ctypes.c_int32),
        ("right", ctypes.c_int32),
        ("bottom", ctypes.c_int32),
    ]
"""


class DiscordFileStat(ctypes.Structure):
    _fields_ = [
        ("filename", ctypes.c_char * 260),
        ("size", ctypes.c_uint64),
        ("last_modified", ctypes.c_uint64),
    ]


class DiscordEntitlement(ctypes.Structure):
    _fields_ = [
        ("id", DiscordSnowflake),
        ("type", ctypes.c_int32),
        ("sku_id", DiscordSnowflake),
    ]


class DiscordSkuPrice(ctypes.Structure):
    _fields_ = [
        ("amount", ctypes.c_uint32),
        ("currency", ctypes.c_char * 16),
    ]


class DiscordSku(ctypes.Structure):
    _fields_ = [
        ("id", DiscordSnowflake),
        ("type", ctypes.c_int32),
        ("name", ctypes.c_char * 256),
        ("price", DiscordSkuPrice),
    ]


class DiscordInputMode(ctypes.Structure):
    _fields_ = [
        ("type", ctypes.c_int32),
        ("shortcut", ctypes.c_char * 256),
    ]


class DiscordUserAchievement(ctypes.Structure):
    _fields_ = [
        ("user_id", DiscordSnowflake),
        ("achievement_id", DiscordSnowflake),
        ("percent_complete", ctypes.c_uint8),
        ("unlocked_at", DiscordDateTime),
    ]


class IDiscordLobbyTransaction(ctypes.Structure):
    pass


IDiscordLobbyTransaction._fields_ = [
    ("set_type", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        ctypes.c_int32
    )),
    ("set_owner", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        DiscordUserId
    )),
    ("set_capacity", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        ctypes.c_uint32
    )),
    ("set_metadata", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        DiscordMetadataKey,
        DiscordMetadataValue
    )),
    ("delete_metadata", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        DiscordMetadataKey
    )),
    ("set_locked", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyTransaction),
        ctypes.c_bool
    )),
]


class IDiscordLobbyMemberTransaction(ctypes.Structure):
    pass


IDiscordLobbyMemberTransaction._fields_ = [
    ("set_metadata", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyMemberTransaction),
        DiscordMetadataKey, DiscordMetadataValue
    )),
    ("delete_metadata", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyMemberTransaction),
        DiscordMetadataKey
    )),
]


class IDiscordLobbySearchQuery(ctypes.Structure):
    pass


IDiscordLobbySearchQuery._fields_ = [
    ("filter", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbySearchQuery),
        DiscordMetadataKey,
        ctypes.c_int32,
        ctypes.c_int32,
        DiscordMetadataValue
    )),
    ("sort", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbySearchQuery),
        DiscordMetadataKey,
        ctypes.c_int32,
        DiscordMetadataValue
    )),
    ("limit", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbySearchQuery),
        ctypes.c_uint32
    )),
    ("distance", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbySearchQuery),
        ctypes.c_int32
    )),
]

IDiscordApplicationEvents = ctypes.c_void_p


class IDiscordApplicationManager(ctypes.Structure):
    pass


IDiscordApplicationManager._fields_ = [
    ("validate_or_exit", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("get_current_locale", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.POINTER(DiscordLocale)
    )),
    ("get_current_branch", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.POINTER(DiscordBranch)
    )),
    ("get_oauth2_token", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordOAuth2Token)
        )
    )),
    ("get_ticket", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.c_char_p
        )
    )),
]


class IDiscordUserEvents(ctypes.Structure):
    _fields_ = [
        ("on_current_user_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p
        )),
    ]


class IDiscordUserManager(ctypes.Structure):
    pass


IDiscordUserManager._fields_ = [
    ("get_current_user", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordUserManager),
        ctypes.POINTER(DiscordUser)
    )),
    ("get_user", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordUserManager),
        DiscordUserId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordUser)
        )
    )),
    ("get_current_user_premium_type", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordUserManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("current_user_has_flag", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordUserManager),
        ctypes.c_int32,
        ctypes.POINTER(ctypes.c_bool)
    )),
]

IDiscordImageEvents = ctypes.c_void_p


class IDiscordImageManager(ctypes.Structure):
    pass


IDiscordImageManager._fields_ = [
    ("fetch", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordImageManager),
        DiscordImageHandle,
        ctypes.c_bool,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            DiscordImageHandle
        )
    )),
    ("get_dimensions", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordImageManager),
        DiscordImageHandle,
        ctypes.POINTER(DiscordImageDimensions)
    )),
    ("get_data", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordImageManager),
        DiscordImageHandle,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32
    )),
]


class IDiscordActivityEvents(ctypes.Structure):
    _fields_ = [
        ("on_activity_join", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_char_p
        )),
        ("on_activity_spectate", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_char_p
        )),
        ("on_activity_join_request", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordUser)
        )),
        ("on_activity_invite", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordUser),
            ctypes.POINTER(DiscordActivity)
        )),
    ]


class IDiscordActivityManager(ctypes.Structure):
    pass


IDiscordActivityManager._fields_ = [
    ("register_command", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordActivityManager),
        ctypes.c_char_p
    )),
    ("register_steam", ctypes.CFUNCTYPE(
        ctypes.c_int32,

        ctypes.POINTER(IDiscordActivityManager),
        ctypes.c_int32
    )),
    ("update_activity", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordActivityManager),
        ctypes.POINTER(DiscordActivity),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_uint32
        )
    )),
    ("clear_activity", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordActivityManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("send_request_reply", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordActivityManager),
        DiscordUserId,
        ctypes.c_int32,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("send_invite", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordActivityManager),
        DiscordUserId,
        ctypes.c_int32,
        ctypes.c_char_p,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("accept_invite", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordActivityManager),
        DiscordUserId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
]


class IDiscordRelationshipEvents(ctypes.Structure):
    _fields_ = [
        ("on_refresh", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p
        )),
        ("on_relationship_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordRelationship)
        )),
    ]


class IDiscordRelationshipManager(ctypes.Structure):
    pass


IDiscordRelationshipManager._fields_ = [
    ("filter", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordRelationshipManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            ctypes.c_bool,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordRelationship)
        )
    )),
    ("count", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordRelationshipManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordRelationshipManager),
        DiscordUserId,
        ctypes.POINTER(DiscordRelationship)
    )),
    ("get_at", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordRelationshipManager),
        ctypes.c_uint32,
        ctypes.POINTER(DiscordRelationship)
    )),
]


class IDiscordLobbyEvents(ctypes.Structure):
    _fields_ = [
        ("on_lobby_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64
        )),
        ("on_lobby_delete", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_uint32
        )),
        ("on_member_connect", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64
        )),
        ("on_member_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64
        )),
        ("on_member_disconnect", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64
        )),
        ("on_lobby_message", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32
        )),
        ("on_speaking", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64,
            ctypes.c_bool
        )),
        ("on_network_message", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int64,
            ctypes.c_int64,
            ctypes.c_uint8,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32
        )),
    ]


class IDiscordLobbyManager(ctypes.Structure):
    pass


IDiscordLobbyManager._fields_ = [
    ("get_lobby_create_transaction", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(ctypes.POINTER(IDiscordLobbyTransaction))
    )),
    ("get_lobby_update_transaction", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(ctypes.POINTER(IDiscordLobbyTransaction))
    )),
    ("get_member_update_transaction", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.POINTER(ctypes.POINTER(IDiscordLobbyMemberTransaction))
    )),
    ("create_lobby", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(IDiscordLobbyTransaction),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordLobby)
        )
    )),
    ("update_lobby", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(IDiscordLobbyTransaction),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("delete_lobby", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("connect_lobby", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordLobbySecret,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordLobby)
        )
    )),
    ("connect_lobby_with_activity_secret", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbySecret,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(DiscordLobby)
        )
    )),
    ("disconnect_lobby", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("get_lobby", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(DiscordLobby)
    )),
    ("get_lobby_activity_secret", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(DiscordLobbySecret)
    )),
    ("get_lobby_metadata_value", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordMetadataKey,
        ctypes.POINTER(DiscordMetadataValue)
    )),
    ("get_lobby_metadata_key", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_int32,
        ctypes.POINTER(DiscordMetadataKey)
    )),
    ("lobby_metadata_count", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("member_count", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get_member_user_id", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_int32,
        ctypes.POINTER(DiscordUserId)
    )),
    ("get_member_user", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.POINTER(DiscordUser)
    )),
    ("get_member_metadata_value", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        DiscordMetadataKey,
        ctypes.POINTER(DiscordMetadataValue)
    )),
    ("get_member_metadata_key", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.c_int32,
        ctypes.POINTER(DiscordMetadataKey)
    )),
    ("member_metadata_count", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("update_member", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.POINTER(IDiscordLobbyMemberTransaction),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("send_lobby_message", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("get_search_query", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(ctypes.POINTER(IDiscordLobbySearchQuery))
    )),
    ("search", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(IDiscordLobbySearchQuery),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("lobby_count", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get_lobby_id", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.c_int32,
        ctypes.POINTER(DiscordLobbyId)
    )),
    ("connect_voice", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("disconnect_voice", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("connect_network", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId
    )),
    ("disconnect_network", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId
    )),
    ("flush_network", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager)
    )),
    ("open_network_channel", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        ctypes.c_uint8,
        ctypes.c_bool
    )),
    ("send_network_message", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordLobbyManager),
        DiscordLobbyId,
        DiscordUserId,
        ctypes.c_uint8,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32
    )),
]


class IDiscordNetworkEvents(ctypes.Structure):
    _fields_ = [
        ("on_message", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            DiscordNetworkPeerId,
            DiscordNetworkChannelId,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32
        )),
        ("on_route_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_char_p
        )),
    ]


class IDiscordNetworkManager(ctypes.Structure):
    pass


IDiscordNetworkManager._fields_ = [
    ("get_peer_id", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordNetworkManager),
        ctypes.POINTER(DiscordNetworkPeerId)
    )),
    ("flush", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager)
    )),
    ("open_peer", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId,
        ctypes.c_char_p
    )),
    ("update_peer", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId,
        ctypes.c_char_p
    )),
    ("close_peer", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId
    )),
    ("open_channel", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId,
        DiscordNetworkChannelId,
        ctypes.c_bool
    )),
    ("close_channel", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId,
        DiscordNetworkChannelId
    )),
    ("send_message", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordNetworkManager),
        DiscordNetworkPeerId,
        DiscordNetworkChannelId,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32
    )),
]


class IDiscordOverlayEvents(ctypes.Structure):
    _fields_ = [
        ("on_toggle", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_bool
        )),
    ]


class IDiscordOverlayManager(ctypes.Structure):
    pass


IDiscordOverlayManager._fields_ = [
    ("is_enabled", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("is_locked", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("set_locked", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.c_bool,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("open_activity_invite", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.c_int32,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("open_guild_invite", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.c_char_p,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("open_voice_settings", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
]

IDiscordStorageEvents = ctypes.c_void_p


class IDiscordStorageManager(ctypes.Structure):
    pass


IDiscordStorageManager._fields_ = [
    ("read", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32,
        ctypes.POINTER(ctypes.c_uint32)
    )),
    ("read_async", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32)
    )),
    ("read_async_partial", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.c_uint64,
        ctypes.c_uint64,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32
        )
    )),
    ("write", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32
    )),
    ("write_async", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_uint8),
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("delete_", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p
    )),
    ("exists", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("count", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("stat", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_char_p,
        ctypes.POINTER(DiscordFileStat)
    )),
    ("stat_at", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.c_int32,
        ctypes.POINTER(DiscordFileStat)
    )),
    ("get_path", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.POINTER(DiscordPath)
    )),
]


class IDiscordStoreEvents(ctypes.Structure):
    _fields_ = [
        ("on_entitlement_create", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordEntitlement)
        )),
        ("on_entitlement_delete", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordEntitlement)
        )),
    ]


class IDiscordStoreManager(ctypes.Structure):
    pass


IDiscordStoreManager._fields_ = [
    ("fetch_skus", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("count_skus", ctypes.CFUNCTYPE(
        None, ctypes.POINTER(IDiscordStoreManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get_sku", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStoreManager),
        DiscordSnowflake,
        ctypes.POINTER(DiscordSku)
    )),
    ("get_sku_at", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.c_int32,
        ctypes.POINTER(DiscordSku)
    )),
    ("fetch_entitlements", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("count_entitlements", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get_entitlement", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStoreManager),
        DiscordSnowflake,
        ctypes.POINTER(DiscordEntitlement)
    )),
    ("get_entitlement_at", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.c_int32,
        ctypes.POINTER(DiscordEntitlement)
    )),
    ("has_sku_entitlement", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordStoreManager),
        DiscordSnowflake,
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("start_purchase", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordStoreManager),
        DiscordSnowflake,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
]


class IDiscordVoiceEvents(ctypes.Structure):
    _fields_ = [
        ("on_settings_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p
        )),
    ]


class IDiscordVoiceManager(ctypes.Structure):
    pass


IDiscordVoiceManager._fields_ = [
    ("get_input_mode", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.POINTER(DiscordInputMode)
    )),
    ("set_input_mode", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordVoiceManager),
        DiscordInputMode,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("is_self_mute", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("set_self_mute", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.c_bool
    )),
    ("is_self_deaf", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("set_self_deaf", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.c_bool
    )),
    ("is_local_mute", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        DiscordSnowflake,
        ctypes.POINTER(ctypes.c_bool)
    )),
    ("set_local_mute", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        DiscordSnowflake,
        ctypes.c_bool
    )),
    ("get_local_volume", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        DiscordSnowflake,
        ctypes.POINTER(ctypes.c_uint8)
    )),
    ("set_local_volume", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordVoiceManager),
        DiscordSnowflake,
        ctypes.c_uint8
    )),
]


class IDiscordAchievementEvents(ctypes.Structure):
    _fields_ = [
        ("on_user_achievement_update", ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.POINTER(DiscordUserAchievement)
        )),
    ]


class IDiscordAchievementManager(ctypes.Structure):
    pass


IDiscordAchievementManager._fields_ = [
    ("set_user_achievement", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordAchievementManager),
        DiscordSnowflake,
        ctypes.c_uint8,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("fetch_user_achievements", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordAchievementManager),
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32
        )
    )),
    ("count_user_achievements", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordAchievementManager),
        ctypes.POINTER(ctypes.c_int32)
    )),
    ("get_user_achievement", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordAchievementManager),
        DiscordSnowflake,
        ctypes.POINTER(DiscordUserAchievement)
    )),
    ("get_user_achievement_at", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordAchievementManager),
        ctypes.c_int32,
        ctypes.POINTER(DiscordUserAchievement)
    )),
]

IDiscordCoreEvents = ctypes.c_void_p


class IDiscordCore(ctypes.Structure):
    pass


IDiscordCore._fields_ = [
    ("destroy", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordCore)
    )),
    ("run_callbacks", ctypes.CFUNCTYPE(
        ctypes.c_int32,
        ctypes.POINTER(IDiscordCore)
    )),
    ("set_log_hook", ctypes.CFUNCTYPE(
        None,
        ctypes.POINTER(IDiscordCore),
        ctypes.c_int32,
        ctypes.c_void_p,
        ctypes.CFUNCTYPE(
            None,
            ctypes.c_void_p,
            ctypes.c_int32,
            ctypes.c_char_p
        )
    )),
    ("get_application_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordApplicationManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_user_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordUserManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_image_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordImageManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_activity_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordActivityManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_relationship_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordRelationshipManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_lobby_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordLobbyManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_network_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordNetworkManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_overlay_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordOverlayManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_storage_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordStorageManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_store_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordStoreManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_voice_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordVoiceManager),
        ctypes.POINTER(IDiscordCore)
    )),
    ("get_achievement_manager", ctypes.CFUNCTYPE(
        ctypes.POINTER(IDiscordAchievementManager),
        ctypes.POINTER(IDiscordCore)
    )),
]


class DiscordCreateParams(ctypes.Structure):
    _fields_ = [
        ("client_id", DiscordClientId),
        ("flags", ctypes.c_uint64),
        ("events", ctypes.POINTER(IDiscordCoreEvents)),
        ("event_data", ctypes.c_void_p),
        ("application_events", ctypes.POINTER(IDiscordApplicationEvents)),
        ("application_version", DiscordVersion),
        ("user_events", ctypes.POINTER(IDiscordUserEvents)),
        ("user_version", DiscordVersion),
        ("image_events", ctypes.POINTER(IDiscordImageEvents)),
        ("image_version", DiscordVersion),
        ("activity_events", ctypes.POINTER(IDiscordActivityEvents)),
        ("activity_version", DiscordVersion),
        ("relationship_events", ctypes.POINTER(IDiscordRelationshipEvents)),
        ("relationship_version", DiscordVersion),
        ("lobby_events", ctypes.POINTER(IDiscordLobbyEvents)),
        ("lobby_version", DiscordVersion),
        ("network_events", ctypes.POINTER(IDiscordNetworkEvents)),
        ("network_version", DiscordVersion),
        ("overlay_events", ctypes.POINTER(IDiscordOverlayEvents)),
        ("overlay_version", DiscordVersion),
        ("storage_events", ctypes.POINTER(IDiscordStorageEvents)),
        ("storage_version", DiscordVersion),
        ("store_events", ctypes.POINTER(IDiscordStoreEvents)),
        ("store_version", DiscordVersion),
        ("voice_events", ctypes.POINTER(IDiscordVoiceEvents)),
        ("voice_version", DiscordVersion),
        ("achievement_events", ctypes.POINTER(IDiscordAchievementEvents)),
        ("achievement_version", DiscordVersion)
    ]


def DiscordCreateParamsSetDefault(params):
    params.application_version = 1
    params.user_version = 1
    params.image_version = 1
    params.activity_version = 1
    params.relationship_version = 1
    params.lobby_version = 1
    params.network_version = 1
    params.overlay_version = 1
    params.storage_version = 1
    params.store_version = 1
    params.voice_version = 1
    params.achievement_version = 1
